//##################################################################################################
//
//  Ceetron Result Calculator Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2019, Ceetron AS
//##################################################################################################

#include "TDLogging.h"

#include "crchlPluginManager.h"
#include "crchlCalculatorInstance.h"
#include "crchlCalculatorMetaData.h"
#include "crchlHostMetaData.h"
#include "crchlInitOptions.h"
#include "crchlError.h"

#include "CRCPlugin.h"
#include "CRCUtils.h"

#include <cstdlib>
#include <cstdio>


//--------------------------------------------------------------------------------------------------
///
//--------------------------------------------------------------------------------------------------
void createDummyMetaData(crchl::HostMetaData* metaData)
{
    metaData->addResultInfo(CRC_RESTYPE_SCALAR, CRCString::fromChar("Velocity magnitude"));
    metaData->addResultInfo(CRC_RESTYPE_SCALAR, CRCString::fromChar("Pressure"));
    metaData->addResultInfo(CRC_RESTYPE_VECTOR, CRCString::fromChar("Velocity"));
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    TD_LOG_INFO("TestDriver starting...");
    TD_LOG_INFO("------------------------------");
    
    const int reqArgCount = 2;
    if (argc != reqArgCount)
    {
        TD_LOG_ERROR("Not enough parameters specified.");
        TD_LOG_INFO("");
        TD_LOG_INFO("Usage: TestDriver <plugin_file_base_name>");
        TD_LOG_INFO("");
        return EXIT_FAILURE;
    }

    // This should be the base file name of the plugin that is to be loaded
    // The base file name should NOT include the file extension (.dll, .so or .dylib)
    std::string pluginBaseFilePath = argv[reqArgCount - 1];
    TD_LOG_INFO("Plugin base file name: " + pluginBaseFilePath);

    crchl::PluginManager* pluginMgr = crchl::PluginManager::instance();
    pluginMgr->setLogger(TDLogging::loggerInstance());

    if (!pluginMgr->loadPlugin(pluginBaseFilePath))
    {
        TD_LOG_ERROR("Failed to load and initialize the plugin: " + pluginBaseFilePath);
        return EXIT_FAILURE;
    }
    TD_LOG_INFO("Plugin successfully initialized");

    const size_t numCalculators = pluginMgr->calculatorCount();
    TD_LOG_INFO(TDLogger::printf("Number of available calculators: %d", static_cast<int>(numCalculators)));

    // List calculators available in plugin
    for (size_t i = 0; i < numCalculators; i++)
    {
        const std::string calculatorIdString = pluginMgr->calculatorId(i);
        TD_LOG_INFO(TDLogger::printf("  calculator(%d) idString: %s", static_cast<int>(i), calculatorIdString.c_str()));
    }
    
    // Create dummy host meta data to allow for initializing the calculators
    crchl::HostMetaData hostMetaData;
    createDummyMetaData(&hostMetaData);

    // Instantiate all the calculators and get info
    for (size_t i = 0; i < numCalculators; ++i)
    {
        const std::string calculatorIdString = pluginMgr->calculatorId(i);

        TD_LOG_INFO("");
        TD_LOG_INFO(TDLogger::printf("Creating instance of calculator: %s", calculatorIdString.c_str()));
        crchl::CalculatorInstance* calculatorInstance = pluginMgr->createCalculatorInstance(i);
        if (!calculatorInstance)
        {
            TD_LOG_ERROR("Could not create calculator instance");
            return EXIT_FAILURE;
        }

        TD_LOG_INFO("Initializing calculator");
        crchl::CalculatorMetaData calcMetaData;
        crchl::Error error;
        const bool initOk = calculatorInstance->initialize(hostMetaData, crchl::InitOptions(), &calcMetaData, &error);
        if (initOk)
        {
            TD_LOG_INFO(std::string("  outputResultType:     ") + CRCUtils::resultTypeToString(calcMetaData.outputResultType()));
            TD_LOG_INFO(std::string("  outputResultName:     ") + std::string(calcMetaData.outputResultName().utf8()));
            TD_LOG_INFO(std::string("  outputResultIdString: ") + std::string(calcMetaData.outputResultIdString().utf8()));

            TD_LOG_INFO("  Required input:  ");

            for (size_t resIdx = 0; resIdx < calcMetaData.inputResultIndices().size(); ++resIdx)
            {
                const CRCResultInfo& resInfo = hostMetaData.resultInfo(calcMetaData.inputResultIndices()[resIdx]);
                
                std::stringstream ss;
                ss << "    " << resIdx << ":  inputResultName:  \"" << resInfo.name().utf8() << "\" Type: " << CRCUtils::resultTypeToString(resInfo.type());
                TD_LOG_INFO(ss.str());
            }
        }
        else
        {
            TD_LOG_INFO("Could not initialize the calculator, host meta data not compatible");
        }

        delete calculatorInstance;
        calculatorInstance = NULL;
    }

    TD_LOG_INFO("");
    TD_LOG_INFO("Shutting down plugin manager...");
    pluginMgr->shutdown();


    TD_LOG_INFO("");
    TD_LOG_INFO("TestDriver finished - SUCCESS!");

    return EXIT_SUCCESS;
}
