//##################################################################################################
//
//  Ceetron Result Calculator Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2019, Ceetron AS
//##################################################################################################

#include "crchlPluginManager.h"
#include "crchlCalculatorInstance.h"
#include "crchlDynamicLibrary.h"
#include "crchlBase.h"
#include "crchlDefaultLogger.h"
#include "crchlStringUtils.h"
#include "crchlAssert.h"

#include "CRCPlugin.h"


#if defined(__GNUC__) && (CRCHL_GCC_VER >= 40500)
CRCHL_GCC_DIAGNOSTIC_IGNORE("-Wpedantic")
#endif

#ifdef _MSC_VER
#pragma warning (disable: 4640)
#endif

// Helper macros for writing log messages to a logger
#define CRCHL_LOG_ERROR(theLogger, theMessage)    if ((theLogger))      { (theLogger)->log(Logger::LL_ERROR, (theMessage)); }
#define CRCHL_LOG_WARNING(theLogger, theMessage)  if ((theLogger))      { (theLogger)->log(Logger::LL_WARNING, (theMessage)); }
#define CRCHL_LOG_INFO(theLogger, theMessage)     if ((theLogger))      { (theLogger)->log(Logger::LL_INFO, (theMessage)); }
#define CRCHL_LOG_DEBUG(theLogger, theMessage)    if ((theLogger))      { (theLogger)->log(Logger::LL_DEBUG, (theMessage)); }


namespace crchl {

#if defined(WIN32)
static const std::string DYNLIB_EXTENSION("dll");
#elif defined(__APPLE__)
static const std::string DYNLIB_EXTENSION("dylib");
#else
static const std::string DYNLIB_EXTENSION("so");
#endif


//==================================================================================================
//
// 
//
//==================================================================================================

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PluginManager::PluginManager()
:   m_logger(new DefaultLogger),
    m_pluginBeingInitialized(NULL)
{
    m_frameworkServices.pluginApiMajorVer    = CRC_API_MAJOR_VER;
    m_frameworkServices.pluginApiMinorVer    = CRC_API_MINOR_VER;
    m_frameworkServices.registerCalculatorFunc = staticHandleCalculatorRegistration;
    m_frameworkServices.logFunc              = staticLogFunction;
    m_frameworkServices.setPluginInfoFunc    = staticHandleSetPluginInfo;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PluginManager::~PluginManager()
{
    CRCHL_LOG_DEBUG(m_logger, "PluginManager::~PluginManager()");

    shutdown();
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::hasCalculatorWithId(const char* calculatorIdString) const
{
    for (size_t i = 0; i < m_calculatorEntryArr.size(); i++)
    {
        if (m_calculatorEntryArr[i].calculatorIdString == calculatorIdString)
        {
            return true;
        }
    }

    return false;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
PluginManager* PluginManager::instance()
{
    static PluginManager sl_pluginManager;
    return &sl_pluginManager;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::shutdown()
{
    CRCHL_LOG_DEBUG(m_logger, "PluginManager::shutdown()\n");

    m_calculatorEntryArr.clear();

    for (size_t i = 0; i < m_pluginEntryArr.size(); i++)
    {
        PluginEntry entry = m_pluginEntryArr[i];
        entry.exitFunc();
        delete entry.dynLib;
    }

    m_pluginEntryArr.clear();
}


//--------------------------------------------------------------------------------------------------
/// Note that the logger will be owned and deleted by the plugin manager
//--------------------------------------------------------------------------------------------------
void PluginManager::setLogger(Logger* logger)
{
    if (m_logger != logger)
    {
        delete m_logger;
        m_logger = logger;
    }
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::loadPlugin(const std::string& pluginBaseFilename)
{
    std::string fullPluginFilename = pluginBaseFilename + "." + DYNLIB_EXTENSION;

    // Make sure we don't load the same plugin twice
    for (size_t i = 0; i < m_pluginEntryArr.size(); i++)
    {
        if (m_pluginEntryArr[i].fullPluginFilename == fullPluginFilename)
        {
            CRCHL_LOG_ERROR(m_logger, "The plugin '" + fullPluginFilename + "' is already loaded");
            return false;
        }
    }

    std::string err;
    DynamicLibrary* dynLib = DynamicLibrary::load(fullPluginFilename.c_str(), &err);
    if (!dynLib)
    {
        CRCHL_LOG_ERROR(m_logger, "Error loading plugin dynamic library: " + err);
        return false;
    }

    // Get the plugin init function
    CRC_PluginInitFunc pluginInitFunc = reinterpret_cast<CRC_PluginInitFunc>(dynLib->getSymbol("initializeResultCalculatorPlugin"));
    if (!pluginInitFunc)
    {
        // Dynamic library missing entry point?
        CRCHL_LOG_ERROR(m_logger, "Could not find initializeResultCalculatorPlugin entry point in plugin");
        delete dynLib;
        return false;
    }

    if (!doInitializationOfPlugin(pluginInitFunc, dynLib, fullPluginFilename))
    {
        CRCHL_LOG_ERROR(m_logger, "Error initializing dynamic plugin");
        delete dynLib;
        return false;
    }

    return true;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::doInitializationOfPlugin(CRC_PluginInitFunc pluginInitFunc, DynamicLibrary* dynLib, std::string fullPluginFilename)
{
    PluginEntry pluginEntry;
    pluginEntry.fullPluginFilename = fullPluginFilename;
    pluginEntry.dynLib = dynLib;

    m_pluginBeingInitialized = &pluginEntry;
    CRC_PluginExitFunc pluginExitFunc = pluginInitFunc(&m_frameworkServices);
    m_pluginBeingInitialized = NULL;

    if (!pluginExitFunc)
    {
        return false;
    }

    pluginEntry.exitFunc = pluginExitFunc;
    m_pluginEntryArr.push_back(pluginEntry);

    CRCHL_LOG_DEBUG(m_logger, "Plugin successfully initialized");
    CRCHL_LOG_DEBUG(m_logger, "  pluginFileName: " + fullPluginFilename);
    
    const size_t numInfoKeys = pluginEntry.pluginInfoKeyValues.size();
    CRCHL_LOG_DEBUG(m_logger, "  pluginInfo(" + StringUtils::toStdString(numInfoKeys) + ")");

    for (std::map<std::string, std::string>::iterator it = pluginEntry.pluginInfoKeyValues.begin(); it != pluginEntry.pluginInfoKeyValues.end(); ++it)
    {
        CRCHL_LOG_DEBUG(m_logger, "    " + it->first + ": " + it->second);
    }

    return true;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
bool PluginManager::initializeStaticPlugin(CRC_PluginInitFunc pluginInitFunc)
{
    CRCHL_ASSERT(pluginInitFunc);

    if (!doInitializationOfPlugin(pluginInitFunc, NULL, std::string()))
    {
        CRCHL_LOG_ERROR(m_logger, "Error initializing static plugin");
        return false;
    }

    return true;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
size_t PluginManager::calculatorCount() const
{
    return m_calculatorEntryArr.size();
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
std::string PluginManager::calculatorId(size_t index) const
{
    return m_calculatorEntryArr[index].calculatorIdString;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
crchl::CalculatorInstance* PluginManager::createCalculatorInstance(size_t index) const
{
    const CalculatorEntry& calcEntry = m_calculatorEntryArr[index];

    CRCHL_ASSERT(calcEntry.createCalculatorFunc);
    CRCResultCalculator* newCalculatorToBeWrapped = calcEntry.createCalculatorFunc();
    if (!newCalculatorToBeWrapped)
    {
        return NULL;
    }

    CRCHL_ASSERT(calcEntry.destroyCalculatorFunc);
    CalculatorInstance* newCalculatorInstance = new CalculatorInstance(calcEntry.calculatorIdString, newCalculatorToBeWrapped, calcEntry.destroyCalculatorFunc);
    return newCalculatorInstance;
}

//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
int PluginManager::handleCalculatorRegistration(const char* calculatorIdString, const CRC_RegisterCalculatorParams* params)
{
    CRCHL_LOG_DEBUG(m_logger, "PluginManager::handleCalculatorRegistration()");
    CRCHL_LOG_DEBUG(m_logger, "  calculatorId: " + StringUtils::toStdString(calculatorIdString));
    CRCHL_LOG_DEBUG(m_logger, "  pluginApiVersion: " + StringUtils::toStdString(params->pluginApiMajorVer) + ", " + StringUtils::toStdString(params->pluginApiMinorVer));

    if (params->pluginApiMajorVer != CRC_API_MAJOR_VER)
    {
        CRCHL_LOG_ERROR(m_logger, "The plugin API Major Version (" + StringUtils::toStdString(params->pluginApiMajorVer) + ") does not match the host's version (" + StringUtils::toStdString(CRC_API_MAJOR_VER) + ").");
        return CRC_RET_FAIL;
    }
    
    if (hasCalculatorWithId(calculatorIdString))
    {
        CRCHL_LOG_ERROR(m_logger, "A calculator with idString '" + StringUtils::toStdString(calculatorIdString) + "' is already present");
        return CRC_RET_FAIL;
    }

    if (!params->createCalculatorFunc)
    {
        CRCHL_LOG_ERROR(m_logger, "No factory function for creating calculator instances was specified (createCalculatorFunc)");
        return CRC_RET_FAIL;
    }

    if (!params->destroyCalculatorFunc)
    {
        CRCHL_LOG_ERROR(m_logger, "No function for destroying calculator instances was specified (destroyCalculatorFunc)");
        return CRC_RET_FAIL;
    }

    CalculatorEntry calcEntry;
    calcEntry.calculatorIdString = calculatorIdString;
    calcEntry.createCalculatorFunc = params->createCalculatorFunc;
    calcEntry.destroyCalculatorFunc = params->destroyCalculatorFunc;
    m_calculatorEntryArr.push_back(calcEntry);

    return CRC_RET_SUCCESS;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
int PluginManager::staticHandleCalculatorRegistration(const char* calculatorIdString, const CRC_RegisterCalculatorParams* params)
{
    PluginManager* mgr = PluginManager::instance();
    return mgr->handleCalculatorRegistration(calculatorIdString, params);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::handleSetPluginInfo(const char* key, const char* value)
{
    if (!m_pluginBeingInitialized)
    {
        CRCHL_LOG_ERROR(m_logger, "Cannot set plugin info outside of plugin initialization");
        return;
    }

    if (!key || !value)
    {
        CRCHL_LOG_ERROR(m_logger, "Cannot set plugin info, either key or value is invalid");
        return;
    }

    m_pluginBeingInitialized->pluginInfoKeyValues[std::string(key)] = value;
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::staticHandleSetPluginInfo(const char* key, const char* value)
{
    PluginManager* mgr = PluginManager::instance();
    return mgr->handleSetPluginInfo(key, value);
}


//--------------------------------------------------------------------------------------------------
/// 
//--------------------------------------------------------------------------------------------------
void PluginManager::staticLogFunction(int level, const char* message)
{
    PluginManager* mgr = PluginManager::instance();
    if (mgr->m_logger)
    {
        Logger::Level loggerLevel = Logger::LL_ERROR;
        if      (level == 1) loggerLevel = Logger::LL_ERROR;
        else if (level == 2) loggerLevel = Logger::LL_WARNING;
        else if (level == 3) loggerLevel = Logger::LL_INFO;
        else if (level == 4) loggerLevel = Logger::LL_DEBUG;

        mgr->m_logger->log(loggerLevel, message);
    }
}

} // crchl

