//##################################################################################################
//
//  Ceetron Result Calculator Framework
//
//  --------------------------------------------------------------------------------------------
//  Copyright (C) 2019, Ceetron AS
//##################################################################################################
#pragma once

#include "CRCBaseTypes.h"
#include "CRCVersion.h"

class CRCResultCalculator;

#define CRC_RET_SUCCESS 1
#define CRC_RET_FAIL    0


extern "C" {

/// Type definition of function pointer for creating new CRCResultCalculator instances
typedef CRCResultCalculator* (*CRC_CreateCalculatorFunc)();

/// Type definition of function pointer for deleting CRCResultCalculator instances
typedef void (*CRC_DestroyCalculatorFunc)(CRCResultCalculator* calcObjPtr);

//==================================================================================================
///
/// \struct CRC_RegisterCalculatorParams
/// 
/// Parameter struct utilized when registering result calculators
//==================================================================================================
struct CRC_RegisterCalculatorParams
{
    crc_int_t                   pluginApiMajorVer;          ///< Major version of the framework plugin API for which this plugin was written
    crc_int_t                   pluginApiMinorVer;          ///< Minor version of the framework plugin API for which this plugin was written
    CRC_CreateCalculatorFunc    createCalculatorFunc;       ///< Pointer to factory function that should be used to create new calculator instances
    CRC_DestroyCalculatorFunc   destroyCalculatorFunc;      ///< Pointer to function that should be used to delete calculator instances
};

/// Type definition of the function pointer for registering result calculators with the plugin manager
typedef int (*CRC_RegisterCalculatorFunc)(const char* calculatorIdString, const CRC_RegisterCalculatorParams* params);


/// Type definition of log function provided by the framework
/// Log levels are 1=error, 2=warning, 3=info, 4=debug
typedef void (*CRC_LogFunc)(int level, const char* message);


/// Type definition of function provided by the framework for setting plugin information in the form of key, value pairs
typedef void (*CRC_SetPluginInfoFunc)(const char* key, const char* value);


//==================================================================================================
///
/// \struct CRC_FrameworkServices
/// 
/// Aggregates the services that the plugin framework provides to the plugin.
/// This includes version of the plugin API, registering result calculators and functions for logging and reporting information on the plugin itself.
/// This struct is passed to each plugin at initialization time.
//==================================================================================================
struct CRC_FrameworkServices
{
    crc_int_t                  pluginApiMajorVer;      ///< Major version of the framework plugin API
    crc_int_t                  pluginApiMinorVer;      ///< Minor version of the framework plugin API
    CRC_RegisterCalculatorFunc registerCalculatorFunc; ///< Function for registering result calculators
    CRC_LogFunc                logFunc;                ///< Function pointer for log function
    CRC_SetPluginInfoFunc      setPluginInfoFunc;      ///< Function pointer for communicating plugin info
};


// Type definition of the plugin exit function pointer (used by the plugin manager to tell plugins to cleanup before being unloaded)
typedef void (*CRC_PluginExitFunc)();


/// Type definition definition of the plugin initialization function pointer (used by plugin manager to initialize plugins)
/// Note that the return type is the CRC_PluginExitFunc (used by plugin manager to tell plugins to cleanup). 
/// If initialization failed for any reason, it must return NULL to let the plugin manger know that the plugin 
/// wasn't initialized properly. 
/// The plugin should use the framework services to register the result calculators it exports as well as communicating log messages.
typedef CRC_PluginExitFunc (*CRC_PluginInitFunc)(const CRC_FrameworkServices* frameworkServices);



// Make sure that CRC_BUILD_PLUGIN_DLL is defined when building the plugin DLL.
#ifdef WIN32
    #ifdef CRC_BUILD_PLUGIN_DLL
        #define CRC_PLUGIN_EXPORT __declspec(dllexport)
    #else
        #define CRC_PLUGIN_EXPORT __declspec(dllimport)
    #endif
#else
    #define CRC_PLUGIN_EXPORT
#endif


/// Named exported entry point into the plugin
/// This definition is required even though the function is loaded from a dynamic library by name and cast to CRC_PluginInitFunc
/// If this declaration is commented out DynamicLibrary::getSymbol() fails
/// The plugin's initialization function MUST be called "initializeResultCalculatorPlugin" and conform to the CRC_PluginInitFunc signature.
extern "C" CRC_PLUGIN_EXPORT CRC_PluginExitFunc initializeResultCalculatorPlugin(const CRC_FrameworkServices* frameworkServices);

} // extern "C"
